#include "Vector.h"
#include <cmath>

Vector::Vector()
{
	data.clear();	
	size = 0;
}

Vector::Vector(int s)
{
	data.clear();
	size = s;
	for (int i=0; i<size; i++)
	{
		data.push_back(0.0f);
	}
}

Vector& Vector::operator=(const Vector &rhs) {

	// Only do assignment if RHS is a different object from this.
	if (this != &rhs) {
		// Deallocate, allocate new space, copy values...
		data.clear();
		this->size = rhs.size;
		for (int i = 0; i<size;i++)
		{
			data.push_back(rhs.data[i]);
		}
	}

	return *this;
}


Vector& Vector::operator+=(const Vector &rhs)
{
	for (int i=0;i<size;i++)
	{
		data[i]+=rhs.data[i];
	}
	return *this;
}

Vector& Vector::operator-=(const Vector &rhs)
{
	for (int i=0;i<size;i++)
	{
		data[i]-=rhs.data[i];
	}
	return *this;
}

Vector& Vector::operator*=(double a)
{
	for (int i=0;i<size;i++)
	{
		data[i]*=a;
	}
	return *this;
}

Vector Vector::operator+(const Vector &rhs)
{
    Vector result = *this;     // Make a copy of myself. 
    result += rhs;            // Use += to add other to the copy.
    return result;              // All done!
}

Vector Vector::operator-(const Vector &rhs)
{
    Vector result = *this;     // Make a copy of myself. 
    result -= rhs;            // Use += to add other to the copy.
    return result;              // All done!
}

double Vector::dot(const Vector& rhs)
{
	double sum = 0;
	for (int i=0; i<size;i++)
	{
		sum+=data[i]*rhs.data[i];
	}
	return sum;
}

Vector operator*(double a,const Vector& rhs)
{
	Vector out(rhs.size);
	for (int i=0;i<rhs.size;i++)
	{
		out.data[i]=a*rhs.data[i];
	}
	return out;
}

double Vector::magnitude()
{
	double sum = 0;
	for (int i=0; i<size; i++)
	{
		sum += data[i]*data[i];
	}
	return sqrt(sum);
}