
#include <iostream>
#include <cmath>
//#include <windows.h>
#include <GL/glut.h>
#include <vector>
#include "Vector.h"
//#include "FletcherReeves.h"

#include <gsl/gsl_multimin.h>
#include <gsl/gsl_blas.h>

#include "directional_minimize.c"

using namespace::std;

extern int updateMax;

bool isPaused = false;

struct Point2D
{
	double x,y;

	Point2D(double a, double b)
	{
		x = a;
		y = b;
	}
};

const int width = 640;
const int height = 480;
vector<Point2D*> userPoints;
double distanceField[2*width][2*height];
double maxDistance = 1.0f;

Point2D startPoint(-1,-1),endPoint(-1,-1),curPoint(-1,-1);
bool pickStart = false;
bool pickEnd = false;
bool isDone = true;
bool isRunning = false;
bool runOptimization = false;

int totalTime = 10;
double m = 1.0f;
Point2D vel(0,0);
Point2D grav(0.0,0.0);
double curTime = 0;

Vector input(totalTime*2);
Vector output(totalTime*2);
Vector actual(totalTime*2);
Point2D temp(0,0);

bool drawMode = false;
double gradX[width][height];
double gradY[width][height];
bool blackHole = false;
double maxGradX;

typedef struct
{
  int iter;
  double step;
  double max_step;
  double tol;
  gsl_vector *x1;
  gsl_vector *dx1;
  gsl_vector *x2;
  double g0norm;
  double pnorm;
  gsl_vector *p;
  gsl_vector *x0;
  gsl_vector *g0;
  gsl_vector *dx0;
  gsl_vector *dg0;
}
rankTwoQuasiNewton_state_t;

static int rankTwoQuasiNewtonAlloc (void *vstate, size_t n) {
  rankTwoQuasiNewton_state_t *state = (rankTwoQuasiNewton_state_t *) vstate;

  state->x1 = gsl_vector_calloc (n);

  if (state->x1 == 0) {
      GSL_ERROR ("failed to allocate space for x1", GSL_ENOMEM);
    }

  state->dx1 = gsl_vector_calloc (n);

  if (state->dx1 == 0) {
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for dx1", GSL_ENOMEM);
    }

  state->x2 = gsl_vector_calloc (n);

  if (state->x2 == 0) {
      gsl_vector_free (state->dx1);
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for x2", GSL_ENOMEM);
    }

  state->p = gsl_vector_calloc (n);

  if (state->p == 0) {
      gsl_vector_free (state->x2);
      gsl_vector_free (state->dx1);
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for p", GSL_ENOMEM);
    }

  state->x0 = gsl_vector_calloc (n);

  if (state->x0 == 0) {
      gsl_vector_free (state->p);
      gsl_vector_free (state->x2);
      gsl_vector_free (state->dx1);
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for x0", GSL_ENOMEM);
    }

  state->g0 = gsl_vector_calloc (n);

  if (state->g0 == 0) {
      gsl_vector_free (state->x0);
      gsl_vector_free (state->p);
      gsl_vector_free (state->x2);
      gsl_vector_free (state->dx1);
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for g0", GSL_ENOMEM);
    }

  state->dx0 = gsl_vector_calloc (n);

  if (state->dx0 == 0) {
      gsl_vector_free (state->g0);
      gsl_vector_free (state->x0);
      gsl_vector_free (state->p);
      gsl_vector_free (state->x2);
      gsl_vector_free (state->dx1);
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for dx0", GSL_ENOMEM);
    }

  state->dg0 = gsl_vector_calloc (n);

  if (state->dg0 == 0) {
      gsl_vector_free (state->dx0);
      gsl_vector_free (state->g0);
      gsl_vector_free (state->x0);
      gsl_vector_free (state->p);
      gsl_vector_free (state->x2);
      gsl_vector_free (state->dx1);
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for dg0", GSL_ENOMEM);
    }

  return GSL_SUCCESS;
}

static int rankTwoQuasiNewtonSet (void *vstate, gsl_multimin_function_fdf * fdf,
                 const gsl_vector * x, double *f, gsl_vector * gradient,
                 double step_size, double tol) {
  rankTwoQuasiNewton_state_t *state = (rankTwoQuasiNewton_state_t *) vstate;

  state->iter = 0;
  state->step = step_size;
  state->max_step = step_size;
  state->tol = tol;

  GSL_MULTIMIN_FN_EVAL_F_DF (fdf, x, f, gradient);

  /* Use the gradient as the initial direction */

  gsl_vector_memcpy (state->x0, x);
  gsl_vector_memcpy (state->p, gradient);
  gsl_vector_memcpy (state->g0, gradient);

  {
    double gnorm = gsl_blas_dnrm2 (gradient);
    state->pnorm = gnorm;
    state->g0norm = gnorm;
  }

  return GSL_SUCCESS;
}

static void rankTwoQuasiNewtonFree (void *vstate) {
  rankTwoQuasiNewton_state_t *state = (rankTwoQuasiNewton_state_t *) vstate;

  gsl_vector_free (state->dg0);
  gsl_vector_free (state->dx0);
  gsl_vector_free (state->g0);
  gsl_vector_free (state->x0);
  gsl_vector_free (state->p);
  gsl_vector_free (state->x2);
  gsl_vector_free (state->dx1);
  gsl_vector_free (state->x1);
}

static int rankTwoQuasiNewtonRestart (void *vstate) {
  rankTwoQuasiNewton_state_t *state = (rankTwoQuasiNewton_state_t *) vstate;

  state->iter = 0;
  return GSL_SUCCESS;
}

/* Take the current location and hessian approxiamation and use them to get 
 * closer to the minimum as well as refining the Hessian Approx. */
static int rankTwoQuasiNewtonIterate (
		     void *vstate, gsl_multimin_function_fdf * fdf,
                     gsl_vector * x, double *f,
                     gsl_vector * gradient, gsl_vector * dx) {
  rankTwoQuasiNewton_state_t *state = (rankTwoQuasiNewton_state_t *) vstate;

  gsl_vector *x1 = state->x1;
  gsl_vector *dx1 = state->dx1;
  gsl_vector *x2 = state->x2;
  gsl_vector *p = state->p;
  gsl_vector *g0 = state->g0;
  gsl_vector *x0 = state->x0;

  double pnorm = state->pnorm;
  double g0norm = state->g0norm;

  double fa = *f, fb, fc;
  double dir;
  double stepa = 0.0, stepb, stepc = state->step, tol = state->tol;

  double g1norm;
  double pg;

  if (pnorm == 0.0 || g0norm == 0.0) {
      gsl_vector_set_zero (dx);
      return GSL_ENOPROG;
    }

  /* Determine direction towards the minimum */

  gsl_blas_ddot (p, gradient, &pg);

  dir = (pg >= 0.0) ? +1.0 : -1.0;

  /* Compute new trial point at x_c= x - step * p, where p is the
     current direction */

  take_step (x, p, stepc, dir / pnorm, x1, dx);

  /* Evaluate function and gradient at new point xc */

  fc = GSL_MULTIMIN_FN_EVAL_F (fdf, x1);

  if (fc < fa) {
      /* Success, reduced the function value */
      state->step = stepc * 2.0;
      *f = fc;
      gsl_vector_memcpy (x, x1);
      GSL_MULTIMIN_FN_EVAL_DF (fdf, x1, gradient);
      return GSL_SUCCESS;
    }

  /* Do a line minimisation in the region (xa,fa) (xc,fc) to find an
     intermediate (xb,fb) satisifying fa > fb < fc.  Choose an initial
     xb based on parabolic interpolation */

  intermediate_point (fdf, x, p, dir / pnorm, pg,
                      stepa, stepc, fa, fc, x1, dx1, gradient, &stepb, &fb);

/* make sure there's a valid step */
  if (stepb == 0.0) {
      return GSL_ENOPROG;
    }

  minimize (fdf, x, p, dir / pnorm,
            stepa, stepb, stepc, fa, fb, fc, tol,
            x1, dx1, x2, dx, gradient, &(state->step), f, &g1norm);

  gsl_vector_memcpy (x, x2);

  /* Choose a new direction for the next step */

  state->iter = (state->iter + 1) % x->size;

  if (state->iter == 0) {
      gsl_vector_memcpy (p, gradient);
      state->pnorm = g1norm;
  } else {
      /* This is the BFGS update: */
      /* p' = g1 - A dx - B dg */
      /* A = - (1+ dg.dg/dx.dg) B + dg.g/dx.dg */
      /* B = dx.g/dx.dg */

      gsl_vector *dx0 = state->dx0;
      gsl_vector *dg0 = state->dg0;

      double dxg, dgg, dxdg, dgnorm, A, B;

      /* dx0 = x - x0 */
      gsl_vector_memcpy (dx0, x);
      gsl_blas_daxpy (-1.0, x0, dx0);

      /* dg0 = g - g0 */
      gsl_vector_memcpy (dg0, gradient);
      gsl_blas_daxpy (-1.0, g0, dg0);

      gsl_blas_ddot (dx0, gradient, &dxg);
      gsl_blas_ddot (dg0, gradient, &dgg);
      gsl_blas_ddot (dx0, dg0, &dxdg);

      dgnorm = gsl_blas_dnrm2 (dg0);

      if (dxdg != 0) {
          B = dxg / dxdg;
          A = -(1.0 + dgnorm * dgnorm / dxdg) * B + dgg / dxdg;
      } else {
          B = 0;
          A = 0; 
      }

      gsl_vector_memcpy (p, gradient);
      gsl_blas_daxpy (-A, dx0, p);
      gsl_blas_daxpy (-B, dg0, p);

      state->pnorm = gsl_blas_dnrm2 (p);
  }

  gsl_vector_memcpy (g0, gradient);
  gsl_vector_memcpy (x0, x);
  state->g0norm = gsl_blas_dnrm2 (g0);

  return GSL_SUCCESS;
}

static const gsl_multimin_fdfminimizer_type rankTwoQuasiNewtonType = {
  "rankTwoQuasiNewton",                /* name */
  sizeof (rankTwoQuasiNewton_state_t),
  &rankTwoQuasiNewtonAlloc,
  &rankTwoQuasiNewtonSet,
  &rankTwoQuasiNewtonIterate,
  &rankTwoQuasiNewtonRestart,
  &rankTwoQuasiNewtonFree
};

const gsl_multimin_fdfminimizer_type
  * gsl_multimin_fdfminimizer_rankTwoQuasiNewton = &rankTwoQuasiNewtonType;

double getFieldX(int i, int j)
{
  
  //return 0;
  if (i <= 0 || j<=0 || i > width-1 || j>height-1) 
    {
      //cout<<"Out of bounds:"<<i<<" "<<j<<endl;
    return 0;
    }
//return 2;

//cout<<"gx:"<<gradX[i][j]<<endl;
  return gradX[i][j];

}

double getFieldY(int i, int j)
{
  if (i <= 0 || j<=0 || i > width-1 || j>height-1) 
    {
      //cout<<"Out of bounds:"<<i<<" "<<j<<endl;
    return 0;
    }

  return gradY[i][j];

}


double getAccY(Vector v, int i)
{
	return ( v.data[2*(i+2) + 1] - 2*v.data[2*(i+1) + 1] + v.data[2*i + 1] );
}

double getAccX(Vector v, int i)
{
	return ( v.data[2*(i+2)] - 2*v.data[2*(i+1)] + v.data[2*i] );
}

double f(Vector v)
{
	double sum = 0;
	sum += pow(v.data[0] - startPoint.x,2); 
	sum += pow(v.data[1] - startPoint.y,2);
	sum += pow(v.data[2*(totalTime-1)] - endPoint.x,2);
	sum += pow(v.data[2*(totalTime-1) + 1] - endPoint.y,2);
	for (int i = 0; i < totalTime - 2; i++)
	{
		// There is no acceleration along X currently
	  double x = v.data[2*i];
	  double y = v.data[2*i + 1];
	  // if (i > 0 && x <= 50 || y <= 50 || x >= width-50 || y >= height-50) return 1e10;
	  sum += pow(m*getAccY(v,i) - m*grav.y - getFieldY(v.data[2*i],v.data[2*i+1]) , 2);
	  sum += pow(m*getAccX(v,i)  - m*grav.x - getFieldX(v.data[2*i],v.data[2*i+1]) , 2);
		// sum += pow(v.data[2*i] - 
	}
	return sum;
}



Vector gradient(Vector v)
{
	Vector g(v.size);
	g.data[0] = 2*(v.data[0] - startPoint.x);
	g.data[1] = 2*(v.data[1] - startPoint.y);
	g.data[2*(totalTime-1)] = 2*(v.data[2*(totalTime-1)] - endPoint.x);
	g.data[2*(totalTime-1) + 1] = 2*(v.data[2*(totalTime-1) + 1] - endPoint.y);


	for (int i = 0; i < totalTime - 2; i++)
	{
		g.data[2*(i+2) + 1] +=  2*m*(m*getAccY(v,i) - m*grav.y - getFieldY(v.data[2*i],v.data[2*i+1]));
		g.data[2*(i+1) + 1] +=  -4*m*(m*getAccY(v,i) - m*grav.y - getFieldY(v.data[2*i],v.data[2*i+1]));
		g.data[2*i + 1] +=  2*m*(m*getAccY(v,i) - m*grav.y - getFieldY(v.data[2*i],v.data[2*i+1]));	

		g.data[2*(i+2)] +=  2*m*(m*getAccX(v,i)- m*grav.x - getFieldX(v.data[2*i],v.data[2*i+1]));
		g.data[2*(i+1)] += -4*m*(m*getAccX(v,i)- m*grav.x - getFieldX(v.data[2*i],v.data[2*i+1] ));
		g.data[2*i ] += 2*m*(m*getAccX(v,i) - m*grav.x - getFieldX(v.data[2*i],v.data[2*i+1]));	
	}

	return g;
}

     double
     my_f (const gsl_vector *v, void *params)
     {
       //cout<<"Computing f...";
       
       
       Vector t(2*totalTime);
       for (int i=0;i<2*totalTime; i++)
	 {
	   t.data[i] = gsl_vector_get(v,i);
	 }
       
      
       //cout<<"Done"<<endl;
       return f(t);
     }
     
     /* The gradient of f, df = (df/dx, df/dy). */
     void 
     my_df (const gsl_vector *v, void *params, 
            gsl_vector *df)
     {

       //cout<<"Computing df...";

       Vector t(2*totalTime);
       for (int i=0;i<2*totalTime; i++)
	 {
	   t.data[i] = gsl_vector_get(v,i);
	 }
       
       Vector g = gradient(t);

       for (int i=0;i<2*totalTime;i++)
	 {
	   gsl_vector_set(df,i, g.data[i]);
	 }

       //cout<<"Done"<<endl;

     }
     
     /* Compute both f and df together. */
     void 
     my_fdf (const gsl_vector *x, void *params, 
             double *f, gsl_vector *df) 
     {
       *f = my_f(x, params); 
       my_df(x, params, df);
     }

void renderScene(void) 
{
	glClearColor(0.0,0.0,0.0,1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLineWidth(0.5f);
	
	glPushMatrix();

	glBegin(GL_POINTS);
	for (int i=1;i<width-1;i++)
	{
		for (int j=1;j<height-1;j++)
		{
		  glColor3f(0,distanceField[i][j]/maxDistance,0);
		  //glColor3f(gradX[i][j]/maxGradX,0,0);

		  //glColor3f((distanceField[i+1][j] - 2*distanceField[i][j] + distanceField[i-1][j])/2*maxDistance,0,0);
			glVertex3f(i,j,0);
		}
	}		
	glEnd();

	glColor3f(0.0,0.0,1.0f);
	glBegin(GL_LINE_STRIP);	
	for (int i=0;i<userPoints.size();i++)
	{
		glVertex3f(userPoints[i]->x,userPoints[i]->y,0.0f);
	}
	glEnd();


	glColor3f(1.0,1.0,0.0);
	glBegin(GL_LINE_STRIP);
	for (int i=0; i< totalTime-1; i++)
	  {
	    glVertex3f(actual.data[2*i],actual.data[2*i+1],0);
	  }
	glEnd();

	glColor3f(1.0,0.0,0.0);
	glBegin(GL_LINE_STRIP);
	for (int i=0; i< totalTime-1; i++)
	  {
	    glVertex3f(output.data[2*i],output.data[2*i+1],0);
	  }
	glEnd();


	int w = 4;
	glBegin(GL_QUADS);
	glColor3f(0.0,1.0,0.0);
	glVertex3f(startPoint.x - w, startPoint.y - w,0);
	glVertex3f(startPoint.x + w, startPoint.y - w,0);
	glVertex3f(startPoint.x + w, startPoint.y + w,0);
	glVertex3f(startPoint.x - w, startPoint.y + w,0);

	glColor3f(1.0,0.0,0.0);
	glVertex3f(endPoint.x - w, endPoint.y - w,0);
	glVertex3f(endPoint.x + w, endPoint.y - w,0);
	glVertex3f(endPoint.x + w, endPoint.y + w,0);
	glVertex3f(endPoint.x - w, endPoint.y + w,0);

	glColor3f(1.0,1.0,1.0);
	glVertex3f(curPoint.x - w, curPoint.y - w,0);
	glVertex3f(curPoint.x + w, curPoint.y - w,0);
	glVertex3f(curPoint.x + w, curPoint.y + w,0);
	glVertex3f(curPoint.x - w, curPoint.y + w,0);

	glEnd();

	glPopMatrix();

	glutSwapBuffers();
	glFlush();
}

void changeSize(int w, int h)
	{

	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	float ratio = 1.0f * w / h;
	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	// Set the clipping volume
	glOrtho(0, width, height, 0.0, -10.0, 10.0);

}

void compute(void)
{
	double delta = 1;
	if (runOptimization)
	{

		input.data[0] = startPoint.x;
		input.data[1] = startPoint.y;
		input.data[2*totalTime - 2] = endPoint.x;
		input.data[2*totalTime - 1] = endPoint.y;

		//fletcherReeves(input, 1e-3, output);

		size_t iter = 0;
		int status;
     
		const gsl_multimin_fdfminimizer_type *T;
		gsl_multimin_fdfminimizer *s;
     		     
		gsl_vector *x;
		gsl_multimin_function_fdf my_func;
     
		my_func.n = 2*totalTime;
		my_func.f = &my_f;
		my_func.df = &my_df;
		my_func.fdf = &my_fdf;
		//my_func.params = &par;
		
		/* Starting point, x = (5,7) */
		gsl_vector_view xView = gsl_vector_view_array (&input.data[0], 2*totalTime);
     
		T = gsl_multimin_fdfminimizer_vector_bfgs;
		//T = gsl_multimin_fdfminimizer_rankTwoQuasiNewton ;
		s = gsl_multimin_fdfminimizer_alloc (T, 2*totalTime);
     
		gsl_multimin_fdfminimizer_set (s, &my_func, &xView.vector, 1e-5, 1e-6);
     
		cout<<"Starting minimization loop.."<<endl;
		do {
		  iter++;
		  status = gsl_multimin_fdfminimizer_iterate (s);
     
		  if (status)
		    break;
		  
		  status = gsl_multimin_test_gradient (s->gradient, 1e-6);
     

		  
		  /* if (status == GSL_SUCCESS)
		     printf ("Minimum found at:\n"); */
		  
		  
		  
     
		  /*
		  printf ("%5d %.5f %.5f %10.5f\n", iter,
			  gsl_vector_get (s->x, 0), 
			  gsl_vector_get (s->x, 1), 
			  s->f);*/
		  
		} while (status == GSL_CONTINUE && iter < 3000);
     
		for (int i=0;i<2*totalTime; i++)
		    {
		      output.data[i] = gsl_vector_get(s->x,i);
		    }
		  
		cout<<"Done with optimization. Number of iterations:"<<iter<<endl;
		

		gsl_multimin_fdfminimizer_free (s);
		
		       
		vel.x = (output.data[2] - output.data[0]);
		vel.y = output.data[3] - output.data[1];

		actual.data[0] = startPoint.x;
		actual.data[1] = startPoint.y;
		actual.data[2] = output.data[2];
		actual.data[3]  = output.data[3];
		Point2D actualVel(0,0);
		actualVel.x = vel.x + getFieldX(actual.data[0],actual.data[1]);
		actualVel.y = vel.y + getFieldY(actual.data[0],actual.data[1]);
		for (int t = 1; t < totalTime-1; t++)
		  {
		    double incX = grav.x + getFieldX(actual.data[2*t],actual.data[2*t+1]);
		    double incY = grav.y + getFieldY(actual.data[2*t],actual.data[2*t+1]);
		    //cout<<"Time:"<<t<<endl;
		    // cout<<"Field at:"<<actual.data[2*t]<<" "<<actual.data[2*t+1]<<" "<<incX<<" "<<incY<<endl;

		    actualVel.x += incX;
		    actualVel.y += incY;

		    //cout<<"Vel:"<<actualVel.x<<" "<<actualVel.y<<endl;
		    actual.data[2*t+ 2] = actual.data[2*t] + actualVel.x;
		    actual.data[2*t + 3] = actual.data[2*t+1] + actualVel.y;

		  }
		//vel.x = 0;
		//vel.y = 100;
		runOptimization = false;		
	}

	
	if (isRunning)
	{
/*		if (curPoint.x < 0 || curPoint.x > width || curPoint.y < 0 || curPoint.y > height)
		{
			isDone = true;
			isRunning = false;
			return;
		} */

		if (curTime >= totalTime ) 
		{
			isDone = true;
			isRunning = false;
			curTime = 0;
			return;
		}
		
		if (curTime > 0)
		{
		temp.x += vel.x * delta;
		temp.y += vel.y * delta;
		vel.y += grav.y*delta;		
		vel.x += grav.x*delta;
		}
		

		curPoint.x = output.data[2*curTime];
		curPoint.y = output.data[2*curTime + 1];

		curTime++;
		
	}

	renderScene();
}

void initializeModel()
{

}

void reset()
{
	isDone = false;
	runOptimization = true;
	curPoint.x = startPoint.x;
	curPoint.y = startPoint.y;

	temp.x = startPoint.x;
	temp.y = startPoint.y;
}

void onMouseClick(int button, int state, int x, int y)
{
	if ( state == GLUT_DOWN ) 
	{
		if (pickStart)
		{
			pickStart = false;
			startPoint.x = x;
			startPoint.y = y;
		}
		else if (pickEnd)
		{
			pickEnd = false;
			endPoint.x = x;
			endPoint.y = y;
		}
		else if (blackHole)
		  {
		    blackHole = false;
		    int radius = 40;
		    int starti = max(0,x - radius);
		    int endi = min(width, x + radius);
		    int startj = max(0,y - radius);
		    int endj = min(height, y + radius);


		    for (int i = starti; i <= endi; i++)
		      {
			for (int j = startj; j <= endj; j++)
			  {
			    if ( pow((double)i-x,2) + pow((double)j-y,2) >radius*radius) continue;
			    gradX[i][j] = -10 *(i - x)/ pow((pow((double)i-x,2) + pow((double)j-y,2)),0.9);
			    gradY[i][j] = -10 *(j - y)/ pow((pow((double)i-x,2) + pow((double)j-y,2)),0.9);

			    if (pow((double)i-x,2) + pow((double)j-y,2) > 0)
			      {
				distanceField[i][j] = 1 / sqrt(pow((double)i-x,2) + pow((double)j-y,2));
			      }
			    //distanceField[i][j] = 100*log(pow((double)i-x,2) + pow((double)j-y,2));
			    else distanceField[i][j] = 0;

			    if (maxDistance < distanceField[i][j])
			      maxDistance = distanceField[i][j];
			  }
		      }
		  }

	}
}

void onMouseMove(int x, int y)
{
  if (pickStart || pickEnd) return;
  if (x == startPoint.x && y == startPoint.y ) return;
  if (x == endPoint.x && y == endPoint.y) return;

  if (!drawMode) return;
	// glutMotionFunc only called when a mouse button is held down
	if (userPoints.size()>0)
	{
		Point2D* lastPoint = userPoints[userPoints.size()-1];
		if ((lastPoint->x - x)*(lastPoint->x - x) + (lastPoint->y - y)*(lastPoint->y - y) < 30) return;
	}
		userPoints.push_back(new Point2D((double)x,(double)y));
}

void cleanUp()
{
	for (int i=0;i<userPoints.size();i++)
	{
		delete userPoints[i];
	}
}


void onKeyPress(unsigned char key, int x, int y)
{
    switch (key) 
    {
        case 27 : 
        case 'q':
			cleanUp();
            exit(0);
            break;

    case 'x':
      if (grav.x == 0.0)
	{
	  grav.x = 2.0;
	}
      else grav.x = 0.0;
      break;

    case 'y':
      if (grav.y == 0.0)
	grav.y = 2.0;
      else grav.y = 0.0;

    case 'c':
      for(int i=0; i<width; i++)
	{
	  for(int j = 0; j<height; j++)
	    {
	      gradX[i][j] = 0;
	      maxDistance = 0;
	      gradY[i][j] = 0;
	      distanceField[i][j] = 0;
	    }
	}
      break;


		case 's':
			pickStart = true;
			break;

		case 'e':
			pickEnd = true;
			break;

    case 'b':
      blackHole = true;
      break;

		case 'r':
			curPoint.x = startPoint.x;
			curPoint.y = startPoint.y;
			temp.x = startPoint.x;
			temp.y = startPoint.y;
			vel.x = (output.data[2] - output.data[0]);
			vel.y = output.data[3] - output.data[1];	
			isRunning = true;
			break;
 
		case ' ':
			isRunning = !isRunning;
			if (isDone)
			{
				// Reset to start and run again
				reset();
				isRunning = true;
			}
			break;


		default :
			break;
    }
    glutPostRedisplay();
}


int main(int argc, char **argv) 
{

  for (int i=0; i<width; i++)
    {
      for (int j=0; j<height; j++)
	{
	  distanceField[i][j] = 0;
	  gradX[i][j] = 0;
	  gradY[i][j] = 0;
	}
    }
	initializeModel();
	// Initialize Glut
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowPosition(100,100);
	glutInitWindowSize(width,height);
	glutCreateWindow("");
	glutIdleFunc(compute);
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);

    glutKeyboardFunc(onKeyPress);
    glutMouseFunc(onMouseClick);
    glutMotionFunc(onMouseMove);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);


	// Start main loop
	glutMainLoop();

	cleanUp();


	return 0;
}
